// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Evaluator.h:	--- interface for clients of the core selector library
//
//	DESCRIPTION:
//		A class Evaluator is defined which gives access to the services
//		provided by the selectcore library.
//
//	CHANGE HISTORY:
//		UL, 1998-01-05: 1.0.0: - first version
//		UL, 1998-05-19: 1.1.0: - new methods: evalLogic()/evalArith()
//								 which evaluate expression in logic/arithmetic
//								 context
//		UL, 1998-10-07: 1.2.0: - minor modifications to support WCS in mask
//								 filter
//		UL, 1999-10-14: 1.3.0: - added isValidExpression() method
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/Evaluator.h,v 1.2 2001/08/10 16:14:56 ulammers Exp $
//

#ifndef EVALUATOR_H
#define EVALUATOR_H

#include <string>
#include <vector>

#include "Selector.h"
#include "TableDescriptor.h"
#include "SelectError.h"

using namespace SelectError;

struct NumericSymbols {
	std::vector<double *>		valuePtr;
	std::vector<std::string>	name;
};

struct StringSymbols {
	std::vector<char *>			textPtr;
	std::vector<unsigned>		maxLength;	//	will consider at most this
											//	many characters
	std::vector<std::string>	name;
};

class Table;

class Evaluator {
public:
	//	exception classes
	class InvalidExpression : public SelectException {
    public:
        InvalidExpression(const std::string &s) : SelectException(s) {}
    };

	class NoArithContext : public SelectException {
    public:
        NoArithContext(const std::string &s) : SelectException(s) {}
    };

	class NoLogicContext : public SelectException {
    public:
        NoLogicContext(const std::string &s) : SelectException(s) {}
    };

	Evaluator(const std::string &expression = "true",
			  const TableDescriptor *tableDescriptor = 0);
		//	construct  Evaluator object with expression (boolean or
		//	arithmetic)

	static bool isValidExpression(const std::string &expression = "true");
		//	check whether the given expression is syntactically correct
		//	and can be used to successfully instantiate an Evaluator
		//	object

	NumericSymbols numericIdentifiers() const;
	StringSymbols stringIdentifiers() const;
		//	return the list of identifiers used in the expression together
		//	with their associated value pointers

	double *rowNumber() const;
		//	return pointer to row number value or 0 if row number identifier
		//	was not used in the expression

	bool isLogic() const;
		//	can the expression be evaluated in logical/arithmetic (true/false)
		//	context?

	bool evalLogic() const;
		//	evaluate expression in logic context

	double evalArith() const;
		//	evaluate expression in arithmetic context

	void setFlagData(const std::string &flagColumnName, unsigned short flagBit)
	const;
		//	register the name of the column containing the selection
		//	status of a row + the bit position in that column; the
		//	SELECTED function needs this

	~Evaluator();
		//	destructor - deletes dynamical tree structure constructed in
		//	constructor

private:
	mutable SelectorGlobals	_selectorGlobals;
	mutable bool			_flagColumnNameSet;
	mutable std::string		_flagColumnName;
};

#endif
