// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	DataFilter.h:	--- the DataFilter class
//
//	DESCRIPTION:
//		A class DataFilter is defined for filtering a data table (Dal notion)
//		or selecting rows of a table with scalar columns according to a given
//		boolean expression.
//
//	CHANGE HISTORY:
//		UL, 1998-01-06: 1.0.0: - first version
//		UL, 1998-06-18: 1.1.0: - a couple of new string constants
//		UL, 1998-06-24: 2.0.0: - better documentation, split generic doIt()
//								 into doFilter() and doSelect() - this breaks
//								 clients
//		UL, 1999-05-12: 2.0.1: - added expandSelectionExpression()
//		UL, 2003-03-12: 2.1.0: - added DataFilter::expression()
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/DataFilter.h,v 1.4 2014/04/15 15:01:34 aibarra Exp $
//

#ifndef DATAFILTER_H
#define DATAFILTER_H

#include <string>
#include <vector>
#include "Evaluator.h"
#include "ColumnGenerator.h"
#if !defined(NON_SAS)
	#include "DalTableDescriptor.h"
#endif

class DataFilter {
public:
	DataFilter();
		//
		//	PURPOSE:
		//		Create a dummy filter object
		//


	DataFilter(const std::string &expression);
		//				 
		//	PURPOSE:
		//		Create a DataFilter object with a boolean expression
		//
		//	ARGUMENTS:
		//		In:
		//			expression:	the boolean selection expression subsequent
		//						doFilter()/doSelect() method invocations will
		//						use	to selected rows or filter the table;
		//						<expression> must adhere to the syntax rules
		//						detailed in User Documentation, otherwise a
		//						fatal error condition is raised


	const std::string &expression() const { return _expression; }
		//
		//	PURPOSE:
		//		Return selection expression of this data filter
		//


	static std::string expandSelectionExpression(const std::string &expression,
												 const Table &table);
		//
		//	PURPOSE:
		//		Expands symbolic references to table attributes by their
		//		string values in a given selection expression. Example:
		//			Table contains:
		//				...
		//				RAWXSQ = '(RAWX-1)**2' / square of RAWX
		//				RAWYSQ = '(RAWY-1)**2' / square of RAWY
		//				DIST   = 'sqrt(#RAWXSQ+#RAWYSQ)' / distance from center
		//				...
		//
		//		expandSelectionExpression("#DIST < 128", table);
		//		will yield:
		//			"sqrt((RAWX-1)**2 + (RAWY-1)**2)"
		//
		//	ARGUMENTS:
		//		In:
		//			expression:	the expression containing symbolic references
		//						to table attributes
		//			table	  : the table containing the referenced attributes
		//
		//	RETURNS:
		//		Selection expression with the attributes references replaced
		//		by their values
		//


	unsigned long doFilter(Table &table) const;
		//
		//	PURPOSE:
		//		Filter a DAL table with the boolean expression this filter
		//		object was instantiated with; a row is removed if
		//		the boolean expression for that row evaluates to false
		//
		//	ARGUMENTS:
		//		InOut:
		//			table:		the table to filter; upon exit, the
		//						rows for which the boolean expression
		//						evaluates to false have been removed, i.e.,
		//						the table has shrunk in size; if the
		//						expression yields true for all rows the
		//						table will not be modified
		//
		//	RETURNS:
		//		the number of selected rows, i.e., the size of the filtered
		//		table
		//


	static const std::string DF_FLAG_COLUMN_NAME;//	name of flag column: FLAG
	static const short	 DF_FLAG_BITPOS	= -1;	 //	bit position; -1 = MSB

	unsigned long doSelect(Table &table,
					   const std::string &flagColumName = DF_FLAG_COLUMN_NAME,
					   short flagBitPos = DF_FLAG_BITPOS) const;
		//
		//	PURPOSE:
		//		Filter a DAL table but never physically remove unselected
		//		rows; instead selected rows get marked by setting a
		//		bit in an integer column of the processed table; for
		//		unselected rows, the corresponding bit in the flag column
		//		will be cleared.
		//
		//	ARGUMENTS:
		//		In:
		//			table:		the table to be processed
		//			flagColumnName:
		//						the name of the integer column used to
		//						store the selection status of a table row;
		//						If the column does not exist, it will be
		//						created with type int8, otherwise the
		//						existing column will be re-used for the
		//						row flagging. If a table with a flag column
		//						which has been added by doSelect() is
		//						subsequently filtered with doFilter() above,
		//						the flag column gets deleted at the end
		//						of the filtering process. In case the flag
		//						column has not been created by doSelect(), the
		//						bit specified by <flagBitPos> is cleared at
		//						the end of doFilter()
		//			flagBitPos:	the position of the bit (Least-Significant-Bit
		//						= 0) used for the row flagging; the default
		//						value of -1 signifies to use the
		//						most-significant-bit (i.e. 31 for int32,
		//						15 for int16, 7 for int8). This minimizes
		//						interference in case the column named via
		//						<flagColumnName> is already in use to store
		//						other information.
		//
		//	RETURNS:
		//		the number of selected rows
		//

	//
	//	the following constant strings are names of attributes that
	//	DataFilter::doSelect() uses to store information about how the
	//	selection status of rows is encoded
	//	
	static const std::string FLAGCOLUMN;// name of string attribute to
										// hold name of flag column: "FLAGCOL"
	static const std::string FLAGBIT;	// name of integer attribute to
										// hold position of selection
										// status bit in column specified
										// in FLAGCOL attribute: "FLAGBIT"
	static const std::string FLAGCREATE;// name of boolean attribute
										// signifying whether the
										// flag column was newly created
										// by doSelect(): "FLAGCRT"

	typedef struct {
			unsigned long	startRow, stopRow;	// from row #startRow until
												// row #stopRow (inclusive)
												// table rows are counted
												// starting form 0
	} RowRange;

	unsigned long doSelect(const Table &table,
						   std::vector<RowRange> &rowRangeVec) const;
		//
		//	PURPOSE:
		//		Evaluate the selection status of each table row but
		//		do not modify the table; return result in the form
		//		of a vector with row ranges
		//
		//	ARGUMENTS:
		//		In:
		//			table:		the table to inspect
		//		Out:
		//			rowRangeVec: result of the selection; each vector element
		//						 represents a continuous block of selected
		//						 rows; the structure `rowRange' consists of
		//						 the two fields `startRow' and `stopRow',
		//						 specifying the first and last row in
		//						 the block of selected rows, respectively;
		//						 rows are counted starting from 0
		//
		//	RETURNS:
		//		the number of selected rows
		//


	~DataFilter();
		//
		//	PURPOSE:
		//		destroy the object constructed with either of the above
		//		constructors

private:
	mutable std::string			_expression;
	mutable DalTableDescriptor	*_dalTableDescriptor;
	mutable Evaluator			*_evaluator;
};

#endif
