// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SelectWCS.h: --- WCSs support
//
//	DESCRIPTION:
//		Instances of class SelectWCS represent transformations to go
//		from integer pixel numbers to real-valued coordinates
//		(and vice-versa) according to World-Coordinate-System
//		Conventions.
//
//	CHANGE HISTORY:
//		UL, 1998-10-06: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Oct  6 15:47:24 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/SelectWCS.h,v 1.3 2001/09/26 13:46:39 ulammers Exp $
//

#ifndef SELECT_WCS_H
#define SELECT_WCS_H

#include <iosfwd>

#include "Dal.h"


class SelectWCS {
public:
	typedef enum { RA_TAN, DEC_TAN } ProjectionType;
	SelectWCS(SelectWCS::ProjectionType projectionType = RA_TAN,
		double refPixel = 0., double refPixelValue = 0.,
		double delta = 1.)
		: _projectionType(projectionType), _refPixel(refPixel),
		  _refPixelValue(refPixelValue), _delta(delta), _keywords(false) {}
		//	default constructor

	SelectWCS(const Array *array, unsigned short axisNr);
		//	construct from array - accessed attributes are: CRPIX, CRVAL, CDELT

	SelectWCS(const Table *table, const std::string &colName);
		//	construct from table - accessed attributes are: TCRPX, TCRVL, TCDLT

	double refPixel() const { return _refPixel; }
		//	What is the refernce pixel?

	double refPixelValue() const { return _refPixelValue; }
		//	What is the value of the refernce pixel?

	double delta() const { return _delta; }
		//	pixel size

	ProjectionType projectionType() const { return _projectionType; }

	bool keywordsPresent() const { return _keywords; }
		//	Parameters read from attributes?

	virtual double operator()(int pixel) const;
		//	apply transformation

	virtual int operator()(double value) const;
		//	inverse of above

	virtual ~SelectWCS() {};
		//	be destructive!

private:
	ProjectionType	_projectionType;
	double			_refPixel, _refPixelValue, _delta;
	bool			_keywords;

	//	private methods
	double paramValue(const Array *, const std::string &, double);
	double paramValue(const Table *table, const std::string &colName,
					  const std::string &attrName, double dftValue = 0.);

	std::string SelectWCSattributeName(const std::string &,
									   unsigned short) const;

};

std::ostream &operator<<(std::ostream &os, const SelectWCS &wcs);

#endif



