// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	TriangulatablePolygon.cc: --- Implementation of TriangulatablePolygon class
//
//	DESCRIPTION:
//		This implements the TriangulatablePolygon class. The code is strongly
//		based and indeed largely identical to the one in package
//		'triangulation' by Atul Narkhede available from
//		ftp://ftp.cs.unc.edu/pub/users/manocha/CODE/Triangulation/triangulation.tar.gz
//		This code in turn is based on the paper "A simple and fast incremental
//		randomized algorithm for computing trapezoidal decompositions and for
//		triangulating polygons" by Raimund Seidel.
//
//		For more information visit the URL:
//		http://www.cs.unc.edu/~dm/CODE/GEM/chapter.html
//
//	CHANGE HISTORY:
//		UL, 2001-08-01: 1.0.0: - first version
//		UL, 2001-09-12: 1.1.0: - bug fix in generate_random_ordering:
//								 max value of randGen was set to 2<<31
//								 but not reset at the end; caused problem
//								 reported in SPR 2153 on the CAL
//		UL, 2002=03-18: 1.1.0: = added #include "Tracer.h"
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug  1 11:40:15 MET DST 2001
//
//	REFERENCES:
//		[1] http://www.cs.unc.edu/~dm/CODE/GEM/chapter.html
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/TriangulatablePolygon.cc,v 1.12 2014/04/15 15:01:34 aibarra Exp $
//

#include <cmath>
#include <cstring> // memset
#include <climits>

#include "Tracer.h"
#include "SelectError.h"
#include "Random.h"
#include "TriangulatablePolygon.h"

SAS_DEBUG(#include <iostream>)

namespace {

//	Node types
enum {
	T_X		= 1,
	T_Y		= 2,
	T_SINK	= 3
};

//	checking whether pt. is inserted
enum {
	FIRSTPT = 1,
	LASTPT	= 2
};

const double TRIA_INF	= double(1<<30);
const double C_EPS		= 1.0e-7;	// tolerance value: Used for making
									// all decisions about collinearity or
									// left/right of segment. Decrease
									// this value if the input points are
									// spaced very close together

//	for merge-direction
enum {
	S_LEFT	= 1,
	S_RIGHT = 2
};

enum {
	ST_VALID	= 1,
	ST_INVALID	= 2
};

//	for splitting trapezoids
enum {
	SP_SIMPLE_LRUP	= 1,
	SP_SIMPLE_LRDN	= 2,
	SP_2UP_2DN		= 3,
	SP_2UP_LEFT		= 4,
	SP_2UP_RIGHT	= 5,
	SP_2DN_LEFT		= 6,
	SP_2DN_RIGHT	= 7,
	SP_NOSPLIT		= -1
};

//	for traverse-direction
enum {
	TR_FROM_UP		= 1,
	TR_FROM_DN		= 2
};

enum {
	TRI_LHS			= 1,
	TRI_RHS			= 2
};


template<class T>
inline T MAX(T a, T b)
{
	return a > b ? a : b;
}


template<class T>
inline T MIN(T a, T b)
{
	return a < b ? a : b;
}


typedef TriangulatablePolygon::point_t	point_t;

inline double CROSS(const point_t &v0, const point_t &v1, const point_t &v2)
{
	return (v1.x - v0.x)*(v2.y - v0.y) - (v1.y - v0.y)*(v2.x - v0.x);
}


inline double DOT(const point_t &v0, const point_t &v1)
{
	return v0.x*v1.x + v0.y*v1.y;
}


inline bool FP_EQUAL(double s, double t)
{
	return std::fabs(s-t) <= C_EPS;
}


inline double log2(double x)
{
	const double logn2 = 0.69314718055994530941;

	return log(x)/logn2;
} // log2()


inline int math_logstar_n(int n)
{
	double v = (double)n;
	int i = 0;
	for (; v>=1.; i++)
		v = log2(v);

	return i-1;
} // math_logstar_n()


inline int math_N(int n, int h)
{
	double v = double(n);

	for (int i=0; i<h; i++)
		v = log2(v);

	return int(ceil(1.*double(n)/v));
} // math_N()


inline bool _equal_to(const point_t *v0, const point_t *v1)
{
	return (FP_EQUAL(v0->y, v1->y) && FP_EQUAL(v0->x, v1->x));
} // _equal_to()


inline bool _greater_than(const point_t *v0, const point_t *v1)
{
	if (v0->y > v1->y + C_EPS)
		return true;
	else if (v0->y < v1->y - C_EPS)
		return false;
	else
		return v0->x > v1->x;
} // _greater_than()


inline bool _greater_than_equal_to(const point_t *v0, const point_t *v1)
{
	if (v0->y > v1->y + C_EPS)
		return true;
	else if (v0->y < v1->y - C_EPS)
		return false;

	return v0->x >= v1->x;
} // _greater_than_equal_to()


inline bool _less_than(const point_t *v0, const point_t *v1)
{
	if (v0->y < v1->y - C_EPS)
		return true;
	else if (v0->y > v1->y + C_EPS)
		return false;

	return v0->x < v1->x;
} // _less_than()


// Return the maximum of the two points into the yval structure
inline void _max(point_t *yval, point_t *v0, point_t *v1)
{
	if (v0->y > v1->y + C_EPS)
		*yval = *v0;
	else if (FP_EQUAL(v0->y, v1->y))
		if (v0->x > v1->x + C_EPS)
			*yval = *v0;
		else
			*yval = *v1;
	else
		*yval = *v1;
} // _max()


// Return the minimum of the two points into the yval structure
inline void _min(point_t *yval, point_t *v0, point_t *v1)
{
	if (v0->y < v1->y - C_EPS)
		*yval = *v0;
	else if (FP_EQUAL(v0->y, v1->y))
		if (v0->x < v1->x)
			*yval = *v0;
		else
			*yval = *v1;
	else
		*yval = *v1;
} // _min()

}

using namespace SelectError;

TriangulatablePolygon::TriangulatablePolygon(
										const std::vector<double> &xCoords,
										const std::vector<double> &yCoords,
										bool pointsScattered,
										bool testSegments)
		throw(IntersectingSegments)
		: Polygon<double>(xCoords, yCoords, xCoords.size()),
		  _triangulated(false), _testSegments(testSegments)
{
	if (pointsScattered)
		_points.simpleClosedPath();
	else {
		_points.antiClockWiseOrder();
							//	triangulation requires vertices to be ordered
							//	anticlockwise
		if (_points.hasIntersectingSegments())
			throw IntersectingSegments();
	}

	//	check if we want to check the segments in the point-inclusion tests
	if (testSegments) {
		const PointList<double> &vertices = this->vertices();
		unsigned N = vertices.size();
		for (unsigned n=0; n<N; ++n)
			_segments.push_back(Line<double>(vertices[n],
											 vertices[n<N-1 ? n+1 : 0]));
	}

	SEGSIZE = xCoords.size() + 10;
	QSIZE = 10*SEGSIZE;
	TRSIZE = 5*SEGSIZE;

	//	allocate memory for common work arrays and initialize
	_qs = new node_t[QSIZE];
	std::memset(_qs, 0, QSIZE * sizeof(node_t));
	_tr = new trap_t[TRSIZE];
	std::memset(_tr, 0, TRSIZE * sizeof(trap_t));
	_seg = new segment_t[SEGSIZE];
	std::memset(_seg, 0, SEGSIZE * sizeof(segment_t));

	SAS_DEBUG(std::cout<<"Pointlist after simpleClosedPath:";
			  std::cout<<_points<<endl;)

	_choose_idx = 1;
	_q_idx = 1;
	_tr_idx = 1;

	triangulation();
} // TriangulatablePolygon::TriangulatablePolygon()


inline bool TriangulatablePolygon::liesOnSegment(const Point<double> &p) const
{
	unsigned N = _segments.size();
	for (unsigned n=0; n<N; ++n)
		if (_segments[n].contains(p))
			return true;

	return false;
} // TriangulatablePolygon::liesOnSegment()


bool TriangulatablePolygon::contains(const Point<double> &p) const
{
	point_t v;

	v.x = p.x();
	v.y = p.y();

	int trnum = locate_endpoint(&v, &v, 1);
	const trap_t *t = &_tr[trnum];

	if (t->state == ST_INVALID || (t->lseg <= 0) || (t->rseg <= 0))
		//	 if any of the above is true the result is false
		//	 unless we want to test the segments; in this case the
		//	 result is the result of this test
		return !_testSegments ? false : liesOnSegment(p);

	int rseg = t->rseg;
	bool res = _greater_than_equal_to(&_seg[rseg].v1, &_seg[rseg].v0);

	if (res)
		return true;	//	the point is surely in
	else	//	test says the point is out but we have not checked the
			//	segments yet; if this is enabled do it now
		if (_testSegments)
			return liesOnSegment(p);

	return false;
} // TriangulatablePolygon::contains()


unsigned TriangulatablePolygon::triangles() const
{
	return _triangles;
} // TriangulatablePolygon::triangles()


const TriangulatablePolygon::TriangleList &
								TriangulatablePolygon::triangulation() const
{
	if (_triangulated)
		return _triangleList;

	_triangleList = new int[QSIZE][3];

	unsigned ncontours = 1;	//	the algorithm can handle polygons with
							//	holes defined by contours; here we are
							//	just interested in the no-hole case

	const PointList<double> &vertices = this->vertices();
	unsigned ccount = 0,
			 i = 1;
	//gcc4.8.2 portability. unsigned definition of cntr returns a norrowing warning.
	long unsigned int cntr[] = { vertices.size() };

	while (ccount < ncontours) {
		unsigned npoints = cntr[ccount],
				 first = i,
				 last = first + npoints - 1;
		for (unsigned j=0; j<npoints; j++, i++) {
			_seg[i].v0.x = vertices[i-1].x();
			_seg[i].v0.y = vertices[i-1].y();

			if (i == last) {
				_seg[i].next = first;
				_seg[i].prev = i-1;
				_seg[i-1].v1 = _seg[i].v0;
			} else if (i == first) {
				_seg[i].next = i+1;
				_seg[i].prev = last;
				_seg[last].v1 = _seg[i].v0;
			} else {
				_seg[i].prev = i-1;
				_seg[i].next = i+1;
				_seg[i-1].v1 = _seg[i].v0;
			}

			_seg[i].is_inserted = false;
		}
		ccount++;
	}

	unsigned n = i-1;
	initialise(n);
	construct_trapezoids(n);

	//	allocate space for arrays and initialize
	_mchain = new monchain_t[TRSIZE];
	std::memset(_mchain, 0, TRSIZE*sizeof(monchain_t));
	_vert = new vertexchain_t[SEGSIZE];
	std::memset(_vert, 0, SEGSIZE*sizeof(vertexchain_t));
	_mon = new int[SEGSIZE];
	std::memset(_mon, 0, SEGSIZE*sizeof(int));
	_visited = new bool[TRSIZE];
	std::memset(_visited, 0, TRSIZE*sizeof(bool));

	unsigned nmonpoly = monotonate_trapezoids(n);
	triangulate_monotone_polygons(n, nmonpoly, _triangleList);

	_triangulated = true;
	_triangles = vertices.size() - 2;

	return _triangleList;
} // TriangulatablePolygon::triangulation()


TriangulatablePolygon::~TriangulatablePolygon()
{
	//	tidy up
	delete [] _qs;
	delete [] _tr;
	delete [] _seg;
	delete [] _triangleList;
	delete [] _permute;

	delete [] _mchain;
	delete [] _vert;
	delete [] _mon;
	delete [] _visited;
} // TriangulatablePolygon::~TriangulatablePolygon()


//	This is query routine which determines which trapezoid does the
//	point v lie in. The return value is the trapezoid number.
//
int TriangulatablePolygon::locate_endpoint(point_t *v,
										   point_t *vo,
										   int r) const
{
	node_t *rptr = &_qs[r];

	switch (rptr->nodetype) {
	case T_SINK:
		return rptr->trnum;

	case T_Y:
		if (_greater_than(v, &rptr->yval))	// above
			return locate_endpoint(v, vo, rptr->right);
		else if (_equal_to(v, &rptr->yval)) // the point is already
											// inserted.
			if (_greater_than(vo, &rptr->yval)) // above
				return locate_endpoint(v, vo, rptr->right);
			else
				return locate_endpoint(v, vo, rptr->left); // below
		else
			return locate_endpoint(v, vo, rptr->left);	// below

	case T_X:
		if (_equal_to(v, &_seg[rptr->segnum].v0) ||
			_equal_to(v, &_seg[rptr->segnum].v1))
			if (FP_EQUAL(v->y, vo->y)) // horizontal segment
				if (vo->x < v->x)
					return locate_endpoint(v, vo, rptr->left); // left
				else
					return locate_endpoint(v, vo, rptr->right); // right
			else if (is_left_of(rptr->segnum, vo))
				return locate_endpoint(v, vo, rptr->left); // left
			else
				return locate_endpoint(v, vo, rptr->right); // right
		else if (is_left_of(rptr->segnum, v))
			return locate_endpoint(v, vo, rptr->left); // left
		else
			return locate_endpoint(v, vo, rptr->right); // right

    default:
		selerrstr<<"Unforeseen condition in function "
				   "TriangulatablePolygon::locate_endpoint"
				<<selerror("ThisShouldNotHappen", 1);
		break;
	}

	return 0;
} // TriangulatablePolygon::locate_endpoint()


//	Retun true if the vertex v is to the left of line segment no.
//	segnum. Takes care of the degenerate cases when both the vertices
//	have the same y--cood, etc.
bool TriangulatablePolygon::is_left_of(int segnum, point_t *v) const
{
	segment_t *s = &_seg[segnum];
	double area;

	if (_greater_than(&s->v1, &s->v0)) // seg. going upwards
		if (FP_EQUAL(s->v1.y, v->y))
			area = v->x < s->v1.x ? 1.0 : -1.0;
		else if (FP_EQUAL(s->v0.y, v->y))
			area = v->x < s->v0.x ? 1.0 : -1.0;
		else
			area = CROSS(s->v0, s->v1, (*v));
	else // v0 > v1
		if (FP_EQUAL(s->v1.y, v->y))
			area = v->x < s->v1.x ? 1.0 : -1.0;
		else if (FP_EQUAL(s->v0.y, v->y))
			area = v->x < s->v0.x ? 1.0 : -1.0;
		else
			area = CROSS(s->v1, s->v0, (*v));

	return area > 0.;
} // TriangulatablePolygon::is_left_of()


void TriangulatablePolygon::initialise(unsigned n) const
{
	for (unsigned i=1; i<=n; i++)
		_seg[i].is_inserted = false;

	generate_random_ordering(n);
} // TriangulatablePolygon::initialise()



void TriangulatablePolygon::generate_random_ordering(unsigned n) const
{
	_permute = new int[QSIZE];

	int st[SEGSIZE];
	for (unsigned i=0; i<=n; i++)
		st[i] = i;

	double oldMax = randGen.max();
	randGen.max(INT_MAX);
	int *p = st;
	for (unsigned i=1; i<=n; i++, p++) {
		int m = int(randGen.rand()) % (n + 1 - i) + 1;
		_permute[i] = p[m];
		if (m != 1)
			p[m] = p[1];
	}
	randGen.max(oldMax);
} // TriangulatablePolygon::generate_random_ordering()


// Return the next segment in the generated random ordering of all the
// segments in S
int TriangulatablePolygon::choose_segment() const
{
	SAS_DEBUG(std::cout<<"choose_segment: "<<_permute[_choose_idx]
					   <<std::endl;)

	return _permute[_choose_idx++];
} // TriangulatablePolygon::choose_segment()


//	Main routine to perform trapezoidation
void TriangulatablePolygon::construct_trapezoids(int nseg) const
{
	// Add the first segment and get the query structure and trapezoid
	// list initialised
	int root = init_query_structure(choose_segment());

	for (int i=1; i<=nseg; i++)
		_seg[i].root0 = _seg[i].root1 = root;

	for (int h=1; h<=math_logstar_n(nseg); h++) {
		for (int i=math_N(nseg, h -1) + 1; i<=math_N(nseg, h); i++)
			add_segment(choose_segment());

		// Find a new root for each of the segment endpoints
		for (int i=1; i<=nseg; i++)
			find_new_roots(i);
	}

	for (int i=math_N(nseg, math_logstar_n(nseg)) + 1; i<=nseg; i++)
		add_segment(choose_segment());
} // TriangulatablePolygon::construct_trapezoids()


// Initilialise the query structure (Q) and the trapezoid table (T)
// when the first segment is added to start the trapezoidation. The
// query-tree starts out with 4 trapezoids, one S-node and 2 Y-nodes
//
//                4
//   -----------------------------------
//  		  |
//  	1	   |        2
//  		    |
//   -----------------------------------
//                3
//
int TriangulatablePolygon::init_query_structure(int segnum) const
{
	segment_t *s = &_seg[segnum];

	int i1 = newnode();
	_qs[i1].nodetype = T_Y;
	_max(&_qs[i1].yval, &s->v0, &s->v1); // root
	int root = i1;

	int i2;
	_qs[i1].right = i2 = newnode();
	_qs[i2].nodetype = T_SINK;
	_qs[i2].parent = i1;

	int i3;
	_qs[i1].left = i3 = newnode();
	_qs[i3].nodetype = T_Y;
	_min(&_qs[i3].yval, &s->v0, &s->v1); // root
	_qs[i3].parent = i1;

	int i4;
	_qs[i3].left = i4 = newnode();
	_qs[i4].nodetype = T_SINK;
	_qs[i4].parent = i3;

	int i5;
	_qs[i3].right = i5 = newnode();
	_qs[i5].nodetype = T_X;
	_qs[i5].segnum = segnum;
	_qs[i5].parent = i3;

	int i6;
	_qs[i5].left = i6 = newnode();
	_qs[i6].nodetype = T_SINK;
	_qs[i6].parent = i5;

	int i7;
	_qs[i5].right = i7 = newnode();
	_qs[i7].nodetype = T_SINK;
	_qs[i7].parent = i5;

	int t1 = newtrap(),		// middle left
		t2 = newtrap(),		// middle right
		t3 = newtrap(),		// bottom-most
		t4 = newtrap();		// topmost

	_tr[t1].hi = _tr[t2].hi = _tr[t4].lo = _qs[i1].yval;
	_tr[t1].lo = _tr[t2].lo = _tr[t3].hi = _qs[i3].yval;
	_tr[t4].hi.y = TRIA_INF;
	_tr[t4].hi.x = TRIA_INF;
	_tr[t3].lo.y = -TRIA_INF;
	_tr[t3].lo.x = -TRIA_INF;
	_tr[t1].rseg = _tr[t2].lseg = segnum;
	_tr[t1].u0 = _tr[t2].u0 = t4;
	_tr[t1].d0 = _tr[t2].d0 = t3;
	_tr[t4].d0 = _tr[t3].u0 = t1;
	_tr[t4].d1 = _tr[t3].u1 = t2;

	_tr[t1].sink = i6;
	_tr[t2].sink = i7;
	_tr[t3].sink = i4;
	_tr[t4].sink = i2;

	_tr[t1].state = _tr[t2].state = ST_VALID;
	_tr[t3].state = _tr[t4].state = ST_VALID;

	_qs[i2].trnum = t4;
	_qs[i4].trnum = t3;
	_qs[i6].trnum = t1;
	_qs[i7].trnum = t2;

	s->is_inserted = true;
	return root;
} // TriangulatablePolygon::init_query_structure()


// Add in the new segment into the trapezoidation and update Q and T
// structures. First locate the two endpoints of the segment in the
// Q-structure. Then start from the topmost trapezoid and go down to
// the  lower trapezoid dividing all the trapezoids in between .
//
void TriangulatablePolygon::add_segment(int segnum) const
{
	segment_t s = _seg[segnum];
	bool is_swapped = false;
	if (_greater_than(&s.v1, &s.v0)) { // Get higher vertex in v0
		point_t tpt = s.v0;
		s.v0 = s.v1;
		s.v1 = tpt;
		int tmp = s.root0;
		s.root0 = s.root1;
		s.root1 = tmp;
		is_swapped = true;
	}

	int tfirst;
	bool tritop = false;
	if (is_swapped ? !inserted(segnum, LASTPT)
				   : !inserted(segnum, FIRSTPT)) { // insert v0 in the tree
		int tmp_d,
			tu = locate_endpoint(&s.v0, &s.v1, s.root0),
			tl = newtrap();		// tl is the new lower trapezoid
		_tr[tl].state = ST_VALID;
		_tr[tl] = _tr[tu];
		_tr[tu].lo.y = _tr[tl].hi.y = s.v0.y;
		_tr[tu].lo.x = _tr[tl].hi.x = s.v0.x;
		_tr[tu].d0 = tl;
		_tr[tu].d1 = 0;
		_tr[tl].u0 = tu;
		_tr[tl].u1 = 0;

		if (((tmp_d = _tr[tl].d0) > 0) && (_tr[tmp_d].u0 == tu))
			_tr[tmp_d].u0 = tl;
		if (((tmp_d = _tr[tl].d0) > 0) && (_tr[tmp_d].u1 == tu))
			_tr[tmp_d].u1 = tl;

		if (((tmp_d = _tr[tl].d1) > 0) && (_tr[tmp_d].u0 == tu))
			_tr[tmp_d].u0 = tl;
		if (((tmp_d = _tr[tl].d1) > 0) && (_tr[tmp_d].u1 == tu))
			_tr[tmp_d].u1 = tl;

		// Now update the query structure and obtain the sinks for the
		// two trapezoids
		int i1 = newnode(),		// Upper trapezoid sink
			i2 = newnode(),		// Lower trapezoid sink
			sk = _tr[tu].sink;

		_qs[sk].nodetype = T_Y;
		_qs[sk].yval = s.v0;
		_qs[sk].segnum = segnum;	// not really reqd ... maybe later
		_qs[sk].left = i2;
		_qs[sk].right = i1;

		_qs[i1].nodetype = T_SINK;
		_qs[i1].trnum = tu;
		_qs[i1].parent = sk;

		_qs[i2].nodetype = T_SINK;
		_qs[i2].trnum = tl;
		_qs[i2].parent = sk;

		_tr[tu].sink = i1;
		_tr[tl].sink = i2;
		tfirst = tl;
	} else {			// v0 already present
						// Get the topmost intersecting trapezoid
		tfirst = locate_endpoint(&s.v0, &s.v1, s.root0);
		tritop = true;
	}

	int tlast;
	bool tribot = false;
	if ((is_swapped) ? !inserted(segnum, FIRSTPT)
					 : !inserted(segnum, LASTPT)) { // insert v1 in the tree
		int tmp_d,
			tu = locate_endpoint(&s.v1, &s.v0, s.root1),
			tl = newtrap();		// tl is the new lower trapezoid

		_tr[tl].state = ST_VALID;
		_tr[tl] = _tr[tu];
		_tr[tu].lo.y = _tr[tl].hi.y = s.v1.y;
		_tr[tu].lo.x = _tr[tl].hi.x = s.v1.x;
		_tr[tu].d0 = tl;
		_tr[tu].d1 = 0;
		_tr[tl].u0 = tu;
		_tr[tl].u1 = 0;

		if (((tmp_d = _tr[tl].d0) > 0) && (_tr[tmp_d].u0 == tu))
			_tr[tmp_d].u0 = tl;
		if (((tmp_d = _tr[tl].d0) > 0) && (_tr[tmp_d].u1 == tu))
			_tr[tmp_d].u1 = tl;

		if (((tmp_d = _tr[tl].d1) > 0) && (_tr[tmp_d].u0 == tu))
			_tr[tmp_d].u0 = tl;
		if (((tmp_d = _tr[tl].d1) > 0) && (_tr[tmp_d].u1 == tu))
			_tr[tmp_d].u1 = tl;

		// Now update the query structure and obtain the sinks for the
		// two trapezoids
		int i1 = newnode(),		// Upper trapezoid sink
			i2 = newnode(),		// Lower trapezoid sink
			sk = _tr[tu].sink;

		_qs[sk].nodetype = T_Y;
		_qs[sk].yval = s.v1;
		_qs[sk].segnum = segnum;	// not really reqd ... maybe later
		_qs[sk].left = i2;
		_qs[sk].right = i1;

		_qs[i1].nodetype = T_SINK;
		_qs[i1].trnum = tu;
		_qs[i1].parent = sk;

		_qs[i2].nodetype = T_SINK;
		_qs[i2].trnum = tl;
		_qs[i2].parent = sk;

		_tr[tu].sink = i1;
		_tr[tl].sink = i2;
		tlast = tu;
	} else {	// v1 already present
				// Get the lowermost intersecting trapezoid
		tlast = locate_endpoint(&s.v1, &s.v0, s.root1);
		tribot = true;
	}

	// Thread the segment into the query tree creating a new X-node
	// First, split all the trapezoids which are intersected by s into
	// two
	int t = tfirst,	// topmost trapezoid
			tfirstr = 0,
			tlastr = 0;
	while ((t > 0) &&
		   _greater_than_equal_to(&_tr[t].lo, &_tr[tlast].lo)) {
		// traverse from top to bot
		int t_sav, tn_sav,
			sk = _tr[t].sink,
			i1 = newnode(),		// left trapezoid sink
			i2 = newnode();		// right trapezoid sink

		_qs[sk].nodetype = T_X;
		_qs[sk].segnum = segnum;
		_qs[sk].left = i1;
		_qs[sk].right = i2;

		_qs[i1].nodetype = T_SINK;	// left trapezoid (use existing one)
		_qs[i1].trnum = t;
		_qs[i1].parent = sk;

		_qs[i2].nodetype = T_SINK;	// right trapezoid (allocate new)
		int tn;
		_qs[i2].trnum = tn = newtrap();
		_tr[tn].state = ST_VALID;
		_qs[i2].parent = sk;

		if (t == tfirst)
			tfirstr = tn;
		if (_equal_to(&_tr[t].lo, &_tr[tlast].lo))
			tlastr = tn;

		_tr[tn] = _tr[t];
		_tr[t].sink = i1;
		_tr[tn].sink = i2;
		t_sav = t;
		tn_sav = tn;

		if ((_tr[t].d0 <= 0) && (_tr[t].d1 <= 0)) { // case cannot arise
			selerrstr<<"Unforeseen condition in function "
					   "TriangulatablePolygon::add_segment"
					 <<selerror("ThisShouldNotHappen", 2);
			break;
		} else if ((_tr[t].d0 > 0) && (_tr[t].d1 <= 0)) {
   			// only one trapezoid below. partition t into two and make the
			// two resulting trapezoids t and tn as the upper neighbours of
			// the sole lower trapezoid

			if ((_tr[t].u0 > 0) && (_tr[t].u1 > 0)) {
				// continuation of a chain from abv.
				if (_tr[t].usave > 0) { // three upper neighbours
					if (_tr[t].uside == S_LEFT) {
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = -1;
						_tr[tn].u1 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
						_tr[_tr[tn].u1].d0 = tn;
					} else {	// intersects in the right
						_tr[tn].u1 = -1;
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = _tr[t].u0;
						_tr[t].u0 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[t].u1].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
					}
					_tr[t].usave = _tr[tn].usave = 0;
				} else {	// No usave.... simple case
					_tr[tn].u0 = _tr[t].u1;
					_tr[t].u1 = _tr[tn].u1 = -1;
					_tr[_tr[tn].u0].d0 = tn;
				}
			} else {	// fresh seg. or upward cusp
				int tmp_u = _tr[t].u0,
					td0, td1;
				if (((td0 = _tr[tmp_u].d0) > 0) &&
					((td1 = _tr[tmp_u].d1) > 0)) { // upward cusp
					if ((_tr[td0].rseg > 0) &&
						!is_left_of(_tr[td0].rseg, &s.v1)) {
						_tr[t].u0 = _tr[t].u1 = _tr[tn].u1 = -1;
						_tr[_tr[tn].u0].d1 = tn;
					} else {	// cusp going leftwards
						_tr[tn].u0 = _tr[tn].u1 = _tr[t].u1 = -1;
						_tr[_tr[t].u0].d0 = t;
					}
				} else { // fresh segment
					_tr[_tr[t].u0].d0 = t;
					_tr[_tr[t].u0].d1 = tn;
				}
			}

			if (FP_EQUAL(_tr[t].lo.y, _tr[tlast].lo.y) &&
				FP_EQUAL(_tr[t].lo.x, _tr[tlast].lo.x) && tribot) {
				// bottom forms a triangle
				int tmptriseg;

				if (is_swapped)
					tmptriseg = _seg[segnum].prev;
				else
					tmptriseg = _seg[segnum].next;

				if ((tmptriseg > 0) && is_left_of(tmptriseg, &s.v0)) {
					// L-R downward cusp
					_tr[_tr[t].d0].u0 = t;
					_tr[tn].d0 = _tr[tn].d1 = -1;
				} else { // R-L downward cusp
					_tr[_tr[tn].d0].u1 = tn;
					_tr[t].d0 = _tr[t].d1 = -1;
				}
			} else {
				if ((_tr[_tr[t].d0].u0 > 0) && (_tr[_tr[t].d0].u1 > 0)) {
					if (_tr[_tr[t].d0].u0 == t) { // passes thru LHS
						_tr[_tr[t].d0].usave = _tr[_tr[t].d0].u1;
						_tr[_tr[t].d0].uside = S_LEFT;
					} else {
						_tr[_tr[t].d0].usave = _tr[_tr[t].d0].u0;
						_tr[_tr[t].d0].uside = S_RIGHT;
					}
				}
				_tr[_tr[t].d0].u0 = t;
				_tr[_tr[t].d0].u1 = tn;
			}
			t = _tr[t].d0;
		} else if ((_tr[t].d0 <= 0) && (_tr[t].d1 > 0)) {
			// Only one trapezoid below
			if ((_tr[t].u0 > 0) && (_tr[t].u1 > 0)) {
				// continuation of a chain from abv.
				if (_tr[t].usave > 0) { // three upper neighbours
					if (_tr[t].uside == S_LEFT) {
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = -1;
						_tr[tn].u1 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
						_tr[_tr[tn].u1].d0 = tn;
					} else { // intersects in the right
						_tr[tn].u1 = -1;
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = _tr[t].u0;
						_tr[t].u0 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[t].u1].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
					}
					_tr[t].usave = _tr[tn].usave = 0;
				} else { // No usave.... simple case
					_tr[tn].u0 = _tr[t].u1;
					_tr[t].u1 = _tr[tn].u1 = -1;
					_tr[_tr[tn].u0].d0 = tn;
				}
			} else {	// fresh seg. or upward cusp
				int tmp_u = _tr[t].u0,
					td0, td1;
				if (((td0 = _tr[tmp_u].d0) > 0) &&
					((td1 = _tr[tmp_u].d1) > 0)) { // upward cusp
					if ((_tr[td0].rseg > 0) &&
						!is_left_of(_tr[td0].rseg, &s.v1)) {
						_tr[t].u0 = _tr[t].u1 = _tr[tn].u1 = -1;
						_tr[_tr[tn].u0].d1 = tn;
					} else {
						_tr[tn].u0 = _tr[tn].u1 = _tr[t].u1 = -1;
						_tr[_tr[t].u0].d0 = t;
					}
				} else { // fresh segment
					_tr[_tr[t].u0].d0 = t;
					_tr[_tr[t].u0].d1 = tn;
				}
			}

			if (FP_EQUAL(_tr[t].lo.y, _tr[tlast].lo.y) &&
				FP_EQUAL(_tr[t].lo.x, _tr[tlast].lo.x) && tribot) {
				// bottom forms a triangle
				int tmptriseg;

				if (is_swapped)
					tmptriseg = _seg[segnum].prev;
				else
					tmptriseg = _seg[segnum].next;

				if ((tmptriseg > 0) && is_left_of(tmptriseg, &s.v0)) {
					// L-R downward cusp
					_tr[_tr[t].d1].u0 = t;
					_tr[tn].d0 = _tr[tn].d1 = -1;
				} else { // R-L downward cusp
					_tr[_tr[tn].d1].u1 = tn;
					_tr[t].d0 = _tr[t].d1 = -1;
				}
			} else {
				if ((_tr[_tr[t].d1].u0 > 0) && (_tr[_tr[t].d1].u1 > 0)) {
					if (_tr[_tr[t].d1].u0 == t) { // passes thru LHS
						_tr[_tr[t].d1].usave = _tr[_tr[t].d1].u1;
						_tr[_tr[t].d1].uside = S_LEFT;
					} else {
						_tr[_tr[t].d1].usave = _tr[_tr[t].d1].u0;
						_tr[_tr[t].d1].uside = S_RIGHT;
					}
				}
				_tr[_tr[t].d1].u0 = t;
				_tr[_tr[t].d1].u1 = tn;
			}

			t = _tr[t].d1;
		} else {
			// two trapezoids below. Find out which one is intersected by
			// this segment and proceed down that one

			// int tmpseg = _tr[_tr[t].d0].rseg;
			double y0, yt;
			point_t tmppt;
			int tnext;
			bool i_d0 = false,
				 i_d1 = false;

			if (FP_EQUAL(_tr[t].lo.y, s.v0.y)) {
				if (_tr[t].lo.x > s.v0.x)
					i_d0 = true;
				else
					i_d1 = true;
			} else {
				tmppt.y = y0 = _tr[t].lo.y;
				yt = (y0 - s.v0.y)/(s.v1.y - s.v0.y);
				tmppt.x = s.v0.x + yt * (s.v1.x - s.v0.x);

				if (_less_than(&tmppt, &_tr[t].lo))
					i_d0 = true;
				else
					i_d1 = true;
			}

			// check continuity from the top so that the lower-neighbour
			// values are properly filled for the upper trapezoid
			if ((_tr[t].u0 > 0) && (_tr[t].u1 > 0)) {
				// continuation of a chain from abv.
				if (_tr[t].usave > 0) { // three upper neighbours
					if (_tr[t].uside == S_LEFT) {
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = -1;
						_tr[tn].u1 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
						_tr[_tr[tn].u1].d0 = tn;
					} else { // intersects in the right
						_tr[tn].u1 = -1;
						_tr[tn].u0 = _tr[t].u1;
						_tr[t].u1 = _tr[t].u0;
						_tr[t].u0 = _tr[t].usave;

						_tr[_tr[t].u0].d0 = t;
						_tr[_tr[t].u1].d0 = t;
						_tr[_tr[tn].u0].d0 = tn;
					}

					_tr[t].usave = _tr[tn].usave = 0;
				} else { // No usave.... simple case
					_tr[tn].u0 = _tr[t].u1;
					_tr[tn].u1 = -1;
					_tr[t].u1 = -1;
					_tr[_tr[tn].u0].d0 = tn;
				}
			} else { // fresh seg. or upward cusp
				int tmp_u = _tr[t].u0,
					td0, td1;
				if (((td0 = _tr[tmp_u].d0) > 0) &&
					((td1 = _tr[tmp_u].d1) > 0)) { // upward cusp
					if ((_tr[td0].rseg > 0) &&
						!is_left_of(_tr[td0].rseg, &s.v1)) {
						_tr[t].u0 = _tr[t].u1 = _tr[tn].u1 = -1;
						_tr[_tr[tn].u0].d1 = tn;
					} else {
						_tr[tn].u0 = _tr[tn].u1 = _tr[t].u1 = -1;
						_tr[_tr[t].u0].d0 = t;
					}
				} else { // fresh segment
					_tr[_tr[t].u0].d0 = t;
					_tr[_tr[t].u0].d1 = tn;
				}
			}

			if (FP_EQUAL(_tr[t].lo.y, _tr[tlast].lo.y) &&
				FP_EQUAL(_tr[t].lo.x, _tr[tlast].lo.x) && tribot) {
				// this case arises only at the lowest trapezoid..
				// tlast, if the lower endpoint of the segment is
				// already inserted in the structure

				_tr[_tr[t].d0].u0 = t;
				_tr[_tr[t].d0].u1 = -1;
				_tr[_tr[t].d1].u0 = tn;
				_tr[_tr[t].d1].u1 = -1;

				_tr[tn].d0 = _tr[t].d1;
				_tr[t].d1 = _tr[tn].d1 = -1;

				tnext = _tr[t].d1;
			} else if (i_d0) { // intersecting d0
				_tr[_tr[t].d0].u0 = t;
				_tr[_tr[t].d0].u1 = tn;
				_tr[_tr[t].d1].u0 = tn;
				_tr[_tr[t].d1].u1 = -1;

				// new code to determine the bottom neighbours of the
				// newly partitioned trapezoid

				_tr[t].d1 = -1;
				tnext = _tr[t].d0;
			} else { // intersecting d1
				_tr[_tr[t].d0].u0 = t;
				_tr[_tr[t].d0].u1 = -1;
				_tr[_tr[t].d1].u0 = t;
				_tr[_tr[t].d1].u1 = tn;

				// new code to determine the bottom neighbours of the
				// newly partitioned trapezoid
				_tr[tn].d0 = _tr[t].d1;
				_tr[tn].d1 = -1;

				tnext = _tr[t].d1;
			}
			t = tnext;
		}
		_tr[t_sav].rseg = _tr[tn_sav].lseg  = segnum;
	} // end-while

	// Now combine those trapezoids which share common segments. We can
	// use the pointers to the parent to connect these together. This
	// works only because all these new trapezoids have been formed
	// due to splitting by the segment, and hence have only one parent

	int tfirstl = tfirst,
		tlastl = tlast;
	merge_trapezoids(segnum, tfirstl, tlastl, S_LEFT);
	merge_trapezoids(segnum, tfirstr, tlastr, S_RIGHT);

	_seg[segnum].is_inserted = true;
} // TriangulatablePolygon::add_segment()



// Thread in the segment into the existing trapezoidation. The
// limiting trapezoids are given by tfirst and tlast (which are the
// trapezoids containing the two endpoints of the segment. Merges all
// possible trapezoids which flank this segment and have been recently
// divided because of its insertion
//
void TriangulatablePolygon::merge_trapezoids(int segnum, int tfirst,
											 int tlast, int side) const
{
	// First merge polys on the LHS
	int t = tfirst;
	while ((t > 0) && _greater_than_equal_to(&_tr[t].lo, &_tr[tlast].lo)) {
		bool cond;
		int tnext;
		if (side == S_LEFT)
			cond = ((((tnext = _tr[t].d0) > 0) &&
						(_tr[tnext].rseg == segnum)) ||
					(((tnext = _tr[t].d1) > 0) &&
						(_tr[tnext].rseg == segnum)));
		else
			cond = ((((tnext = _tr[t].d0) > 0) &&
						(_tr[tnext].lseg == segnum)) ||
					(((tnext = _tr[t].d1) > 0) &&
						(_tr[tnext].lseg == segnum)));

		if (cond) {
			if ((_tr[t].lseg == _tr[tnext].lseg) &&
				(_tr[t].rseg == _tr[tnext].rseg)) { // good neighbours
				// Use the upper node as the new node i.e. t
				int ptnext = _qs[_tr[tnext].sink].parent;
				if (_qs[ptnext].left == _tr[tnext].sink)
					_qs[ptnext].left = _tr[t].sink;
				else
					_qs[ptnext].right = _tr[t].sink;	// redirect parent

				// Change the upper neighbours of the lower trapezoids
				if ((_tr[t].d0 = _tr[tnext].d0) > 0) // AI gcc4 porting ambiguous 'else'
				{
					if (_tr[_tr[t].d0].u0 == tnext)
						_tr[_tr[t].d0].u0 = t;
					else if (_tr[_tr[t].d0].u1 == tnext)
						_tr[_tr[t].d0].u1 = t;
				}

				if ((_tr[t].d1 = _tr[tnext].d1) > 0) // AI gcc4 porting ambiguous 'else'
				{
					if (_tr[_tr[t].d1].u0 == tnext)
						_tr[_tr[t].d1].u0 = t;
					else if (_tr[_tr[t].d1].u1 == tnext)
						_tr[_tr[t].d1].u1 = t;
				}
				_tr[t].lo = _tr[tnext].lo;
				_tr[tnext].state = ST_INVALID; // invalidate the lower
												  // trapezium
			} else // not good neighbours
				t = tnext;
		} else // do not satisfy the outer if
			t = tnext;
	} // end-while
}


// Update the roots stored for each of the endpoints of the segment.
// This is done to speed up the location-query for the endpoint when
// the segment is inserted into the trapezoidation subsequently
//
void TriangulatablePolygon::find_new_roots(int segnum) const
{
	segment_t *s = &_seg[segnum];

	if (s->is_inserted)
		return;

	s->root0 = locate_endpoint(&s->v0, &s->v1, s->root0);
	s->root0 = _tr[s->root0].sink;

	s->root1 = locate_endpoint(&s->v1, &s->v0, s->root1);
	s->root1 = _tr[s->root1].sink;
} // TriangulatablePolygon::find_new_roots()


int TriangulatablePolygon::newnode() const
{
	if (_q_idx < int(QSIZE))
		return _q_idx++;
	else
		selerrstr<<"Query table overflow QSIZE ("<<QSIZE<<") is too small."
				 <<selerror("TableOverflow", 1);

	return -1;
} // TriangulatablePolygon::newnode()


// Return a free trapezoid
int TriangulatablePolygon::newtrap() const
{
	if (_tr_idx < int(TRSIZE)) {
		_tr[_tr_idx].lseg = -1;
		_tr[_tr_idx].rseg = -1;
		_tr[_tr_idx].state = ST_VALID;
		return _tr_idx++;
	} else
		selerrstr<<"Trapezoid-table table overflow TRSIZE ("<<TRSIZE
				 <<") is too small."
				 <<selerror("TableOverflow", 2);

	return -1;
} // TriangulatablePolygon::newtrap()


// Returns true if the corresponding endpoint of the given segment is
// already inserted into the segment tree. Use the simple test of
// whether the segment which shares this endpoint is already inserted
bool TriangulatablePolygon::inserted(int segnum, int whichpt) const
{
	if (whichpt == FIRSTPT)
		return _seg[_seg[segnum].prev].is_inserted;

	return _seg[_seg[segnum].next].is_inserted;
} // TriangulatablePolygon::inserted()


//
// ---( monotone )-------------------------------------------------------------
//
namespace {

inline double CROSS_SINE(const point_t &v0, const point_t &v1)
{
	return v0.x*v1.y - v1.x*v0.y;
}


inline double LENGTH(const point_t &v0)
{
	return std::sqrt(v0.x*v0.x + v0.y*v0.y);
}

inline double get_angle(const point_t *vp0, const point_t *vpnext,
						const point_t *vp1)
{
	point_t v0, v1;

	v0.x = vpnext->x - vp0->x;
	v0.y = vpnext->y - vp0->y;

	v1.x = vp1->x - vp0->x;
	v1.y = vp1->y - vp0->y;

	if (CROSS_SINE(v0, v1) >= 0)	// sine is positive
		return DOT(v0, v1)/LENGTH(v0)/LENGTH(v1);
	else
		return (-DOT(v0, v1)/LENGTH(v0)/LENGTH(v1) - 2.);
} // get_angle()

}

// Function returns true if the trapezoid lies inside the polygon
bool TriangulatablePolygon::inside_polygon(trap_t *t) const
{
	int rseg = t->rseg;

	if (t->state == ST_INVALID)
		return false;

	if ((t->lseg <= 0) || (t->rseg <= 0))
		return false;

	if (((t->u0 <= 0) && (t->u1 <= 0)) ||
		((t->d0 <= 0) && (t->d1 <= 0))) // triangle
		return (_greater_than(&_seg[rseg].v1, &_seg[rseg].v0));

	return false;
} // TriangulatablePolygon::inside_polygon()


// return a new mon structure from the table
int TriangulatablePolygon::newmon() const
{
	return ++_mon_idx;
} // TriangulatablePolygon::newmon()


// return a new chain element from the table
int TriangulatablePolygon::new_chain_element() const
{
	return ++_chain_idx;
} // TriangulatablePolygon::new_chain_element()


// (v0, v1) is the new diagonal to be added to the polygon. Find which
// chain to use and return the positions of v0 and v1 in p and q
void TriangulatablePolygon::get_vertex_positions(int v0, int v1,
												 int *ip, int *iq) const
{
	vertexchain_t *vp0 = &_vert[v0],
				  *vp1 = &_vert[v1];

	// p is identified as follows. Scan from (v0, v1) rightwards till
	// you hit the first segment starting from v0. That chain is the
	// chain of our interest
	double angle = -4.;
	int tp = 0;
	for (unsigned i=0; i<4; i++) {
		if (vp0->vnext[i] <= 0)
			continue;

		double temp;
		if ((temp = get_angle(&vp0->pt, &(_vert[vp0->vnext[i]].pt),
			&vp1->pt)) > angle) {
			angle = temp;
			tp = i;
		}
	}
	*ip = tp;

	// Do similar actions for q
	angle = -4.;
	int tq = 0;
	for (unsigned i=0; i<4; i++) {
		if (vp1->vnext[i] <= 0)
			continue;
		double temp;
		if ((temp = get_angle(&vp1->pt, &(_vert[vp1->vnext[i]].pt),
			&vp0->pt)) > angle) {
			angle = temp;
			tq = i;
		}
	}
	*iq = tq;
} // TriangulatablePolygon::get_vertex_positions()


// v0 and v1 are specified in anti-clockwise order with respect to
// the current monotone polygon mcur. Split the current polygon into
// two polygons using the diagonal (v0, v1)
//
int TriangulatablePolygon::make_new_monotone_poly(int mcur, int v0, int v1)
const
{
	int mnew = newmon();

	vertexchain_t *vp0 = &_vert[v0],
				  *vp1 = &_vert[v1];

	int ip, iq;
	get_vertex_positions(v0, v1, &ip, &iq);
	int p = vp0->vpos[ip],
		q = vp1->vpos[iq];

	// At this stage, we have got the positions of v0 and v1 in the
	// desired chain. Now modify the linked lists
	int i = new_chain_element(),	// for the new list
		j = new_chain_element();

	_mchain[i].vnum = v0;
	_mchain[j].vnum = v1;

	_mchain[i].next = _mchain[p].next;
	_mchain[_mchain[p].next].prev = i;
	_mchain[i].prev = j;
	_mchain[j].next = i;
	_mchain[j].prev = _mchain[q].prev;
	_mchain[_mchain[q].prev].next = j;

	_mchain[p].next = q;
	_mchain[q].prev = p;

	int nf0 = vp0->nextfree,
		nf1 = vp1->nextfree;

	vp0->vnext[ip] = v1;
	vp0->vpos[nf0] = i;
	vp0->vnext[nf0] = _mchain[_mchain[i].next].vnum;
	vp1->vpos[nf1] = j;
	vp1->vnext[nf1] = v0;

	vp0->nextfree++;
	vp1->nextfree++;

	SAS_DEBUG(std::cout<<"make_poly: mcur = "<<mcur
					   <<", (v0, v1) = ("<<v0<<", "<<v1<<")\n";
			  std::cout<<"next posns = (p, q) = ("<<p<<", "<<q<<")\n";)

	_mon[mcur] = p;
	_mon[mnew] = i;

	return mnew;
} // TriangulatablePolygon::make_new_monotone_poly()


// Main routine to get monotone polygons from the trapezoidation of
// the polygon.
unsigned TriangulatablePolygon::monotonate_trapezoids(unsigned n) const
{
	// First locate a trapezoid which lies inside the polygon
	// and which is triangular
	unsigned i;
	for (i=0; i<TRSIZE; i++)
		if (inside_polygon(&_tr[i]))
			break;

	int tr_start = i;

	// Initialise the mon data-structure and start spanning all the
	// trapezoids within the polygon

#if 0
	for (unsigned i=1; i<=n; i++) {
		_mchain[i].prev = i - 1;
		_mchain[i].next = i + 1;
		_mchain[i].vnum = i;
		_vert[i].pt = _seg[i].v0;
		_vert[i].vnext[0] = i + 1;	// next vertex
		_vert[i].vpos[0] = i;	// locn. of next vertex
		_vert[i].nextfree = 1;
	}
	_mchain[1].prev = n;
	_mchain[n].next = 1;
	_vertn].vnext[0] = 1;
	_vertn].vpos[0] = n;
	_chain_idx = n;
	_mon_idx = 0;
	_mon[0] = 1; // position of any vertex in the first chain

#else

	for (unsigned i=1; i<=n; i++) {
		_mchain[i].prev = _seg[i].prev;
		_mchain[i].next = _seg[i].next;
		_mchain[i].vnum = i;
		_vert[i].pt = _seg[i].v0;
		_vert[i].vnext[0] = _seg[i].next; // next vertex
		_vert[i].vpos[0] = i;	// locn. of next vertex
		_vert[i].nextfree = 1;
	}
	_chain_idx = n;
	_mon_idx = 0;
	_mon[0] = 1; // position of any vertex in the first chain

#endif

	// traverse the polygon
	if (_tr[tr_start].u0 > 0)
		traverse_polygon(0, tr_start, _tr[tr_start].u0, TR_FROM_UP);
	else if (_tr[tr_start].d0 > 0)
		traverse_polygon(0, tr_start, _tr[tr_start].d0, TR_FROM_DN);

	// return the number of polygons created
	return newmon();
} // TriangulatablePolygon::monotonate_trapezoids()


// recursively visit all the trapezoids
int TriangulatablePolygon::traverse_polygon(int mcur, int trnum, int from,
										   int dir) const
{
	trap_t *t = &_tr[trnum];
	bool do_switch = false;

	if ((trnum <= 0) || _visited[trnum])
		return 0;

	_visited[trnum] = true;

	// We have much more information available here.
	// rseg: goes upwards
	// lseg: goes downwards

	// Initially assume that dir = TR_FROM_DN (from the left)
	// Switch v0 and v1 if necessary afterwards

	// special cases for triangles with cusps at the opposite ends.
	// take care of this first
	int mnew = 0, retval = 0, v0 = 0, v1 = 0;
	if ((t->u0 <= 0) && (t->u1 <= 0)) {
		if ((t->d0 > 0) && (t->d1 > 0)) { // downward opening triangle
			v0 = _tr[t->d1].lseg;
			v1 = t->lseg;
			if (from == t->d1) {
				do_switch = true;
				mnew = make_new_monotone_poly(mcur, v1, v0);
				traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
				traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
			} else {
				mnew = make_new_monotone_poly(mcur, v0, v1);
				traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
				traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
			}
		} else {
			retval = SP_NOSPLIT;	// Just traverse all neighbours
			traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
			traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
			traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
			traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
		}
	} else if ((t->d0 <= 0) && (t->d1 <= 0)) {
		if ((t->u0 > 0) && (t->u1 > 0)) { // upward opening triangle
			v0 = t->rseg;
			v1 = _tr[t->u0].rseg;
			if (from == t->u1) {
				do_switch = true;
				mnew = make_new_monotone_poly(mcur, v1, v0);
				traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
				traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
			} else {
				mnew = make_new_monotone_poly(mcur, v0, v1);
				traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
				traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
			}
		} else {
			retval = SP_NOSPLIT;	// Just traverse all neighbours
			traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
			traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
			traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
			traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
		}
	} else if ((t->u0 > 0) && (t->u1 > 0)) {
		if ((t->d0 > 0) && (t->d1 > 0)) { // downward + upward cusps
			v0 = _tr[t->d1].lseg;
			v1 = _tr[t->u0].rseg;
			retval = SP_2UP_2DN;
			if (((dir == TR_FROM_DN) && (t->d1 == from)) ||
				((dir == TR_FROM_UP) && (t->u1 == from))) {
				do_switch = true;
				mnew = make_new_monotone_poly(mcur, v1, v0);
				traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
				traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
				traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
				traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
			} else {
				mnew = make_new_monotone_poly(mcur, v0, v1);
				traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
				traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
				traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
				traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
			}
		} else { // only downward cusp
			if (_equal_to(&t->lo, &_seg[t->lseg].v1)) {
				v0 = _tr[t->u0].rseg;
				v1 = _seg[t->lseg].next;

				retval = SP_2UP_LEFT;
				if ((dir == TR_FROM_UP) && (t->u0 == from)) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
				}
			} else {
				v0 = t->rseg;
				v1 = _tr[t->u0].rseg;
				retval = SP_2UP_RIGHT;
				if ((dir == TR_FROM_UP) && (t->u1 == from)) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
				}
			}
		}
	} else if ((t->u0 > 0) || (t->u1 > 0)) { // no downward cusp
		if ((t->d0 > 0) && (t->d1 > 0)) { // only upward cusp
			if (_equal_to(&t->hi, &_seg[t->lseg].v0)) {
				v0 = _tr[t->d1].lseg;
				v1 = t->lseg;
				retval = SP_2DN_LEFT;
				if (!((dir == TR_FROM_DN) && (t->d0 == from))) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
				}
			} else {
				v0 = _tr[t->d1].lseg;
				v1 = _seg[t->rseg].next;

				retval = SP_2DN_RIGHT;
				if ((dir == TR_FROM_DN) && (t->d1 == from)) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
				}
			}
		} else { // no cusp
			if (_equal_to(&t->hi, &_seg[t->lseg].v0) &&
				_equal_to(&t->lo, &_seg[t->rseg].v0)) {
				v0 = t->rseg;
				v1 = t->lseg;
				retval = SP_SIMPLE_LRDN;
				if (dir == TR_FROM_UP) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
				}
			} else if (_equal_to(&t->hi, &_seg[t->rseg].v1) &&
					   _equal_to(&t->lo, &_seg[t->lseg].v1)) {
				v0 = _seg[t->rseg].next;
				v1 = _seg[t->lseg].next;

				retval = SP_SIMPLE_LRUP;
				if (dir == TR_FROM_UP) {
					do_switch = true;
					mnew = make_new_monotone_poly(mcur, v1, v0);
					traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->d0, trnum, TR_FROM_UP);
				} else {
					mnew = make_new_monotone_poly(mcur, v0, v1);
					traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
					traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
					traverse_polygon(mnew, t->u0, trnum, TR_FROM_DN);
					traverse_polygon(mnew, t->u1, trnum, TR_FROM_DN);
				}
			} else { // no split possible
				retval = SP_NOSPLIT;
				traverse_polygon(mcur, t->u0, trnum, TR_FROM_DN);
				traverse_polygon(mcur, t->d0, trnum, TR_FROM_UP);
				traverse_polygon(mcur, t->u1, trnum, TR_FROM_DN);
				traverse_polygon(mcur, t->d1, trnum, TR_FROM_UP);
			}
		}
	}

	return retval;
} // TriangulatablePolygon::traverse_polygon()


// For each monotone polygon, find the ymax and ymin (to determine the
// two y-monotone chains) and pass on this monotone polygon for greedy
// triangulation.
// Take care not to triangulate duplicate monotone polygons
int TriangulatablePolygon::triangulate_monotone_polygons(unsigned nvert,
														 unsigned nmonpoly,
														 TriangleList op) const
{
	SAS_DEBUG(
		for (unsigned i=0; i<nmonpoly; i++) {
			std::cout<<"\n\nPolygon "<<i<<": ";
			int vfirst = _mchain[_mon[i]].vnum;
			int p = _mchain[_mon[i]].next;
			std::cout<<_mchain[_mon[i]].vnum<<' ';
			while (_mchain[p].vnum != vfirst) {
				std::cout<<_mchain[p].vnum<<' ';
				p = _mchain[p].next;
			}
		}
		std::cout<<std::endl;)

	_op_idx = 0;
	for (unsigned i=0; i<nmonpoly; i++) {
		int vcount = 1;
		bool processed = false;
		int vfirst = _mchain[_mon[i]].vnum;
		point_t ymin = _vert[vfirst].pt,
				ymax = ymin;
		int posmin = _mon[i],
			posmax = posmin;
		_mchain[_mon[i]].marked = true;
		int p = _mchain[_mon[i]].next,
			v;
		while ((v = _mchain[p].vnum) != vfirst) {
			if (_mchain[p].marked) {
				processed = true;
					break;	// break from while
			} else
				_mchain[p].marked = true;

			if (_greater_than(&_vert[v].pt, &ymax)) {
				ymax = _vert[v].pt;
				posmax = p;
			}
			if (_less_than(&_vert[v].pt, &ymin)) {
				ymin = _vert[v].pt;
				posmin = p;
			}
			p = _mchain[p].next;
			vcount++;
		}

		if (processed)		// Go to next polygon
			continue;

		if (vcount == 3) {	// already a triangle
			op[_op_idx][0] = _mchain[p].vnum;
			op[_op_idx][1] = _mchain[_mchain[p].next].vnum;
			op[_op_idx][2] = _mchain[_mchain[p].prev].vnum;
			_op_idx++;
		} else {			// triangulate the polygon
			v = _mchain[_mchain[posmax].next].vnum;
			if (_equal_to(&_vert[v].pt, &ymin)) { // LHS is a single line
				triangulate_single_polygon(nvert, posmax, TRI_LHS, op);
			} else
			triangulate_single_polygon(nvert, posmax, TRI_RHS, op);
		}
	}

	SAS_DEBUG(
		for (unsigned i=0; i<_op_idx; i++)
			std::cout<<"tri #"<<i<<": ("<<op[i][0]<<", "<<op[i][1]<<", "
					 <<op[i][1]<<")\n";)

	return _op_idx;
} // TriangulatablePolygon::triangulate_monotone_polygons()


// A greedy corner-cutting algorithm to triangulate a y-monotone
// polygon in O(n) time.
// Joseph O-Rourke, Computational Geometry in C.
//
int TriangulatablePolygon::triangulate_single_polygon(unsigned nvert,
													  unsigned posmax,
													  unsigned side,
													  TriangleList op) const
{
	int rc[SEGSIZE],
		ri = 0,	// reflex chain
		tmp, v, endv, vpos;

	if (side == TRI_RHS) {	// RHS segment is a single segment
		rc[0] = _mchain[posmax].vnum;
		tmp = _mchain[posmax].next;
		rc[1] = _mchain[tmp].vnum;
		ri = 1;

		vpos = _mchain[tmp].next;
		v = _mchain[vpos].vnum;

		if ((endv = _mchain[_mchain[posmax].prev].vnum) == 0)
			endv = nvert;
	} else {	// LHS is a single segment
		tmp = _mchain[posmax].next;
		rc[0] = _mchain[tmp].vnum;
		tmp = _mchain[tmp].next;
		rc[1] = _mchain[tmp].vnum;
		ri = 1;

		vpos = _mchain[tmp].next;
		v = _mchain[vpos].vnum;

		endv = _mchain[posmax].vnum;
	}

	while ((v != endv) || (ri > 1)) {
		if (ri > 0) {	// reflex chain is non-empty
			if (CROSS(_vert[v].pt, _vert[rc[ri - 1]].pt, _vert[rc[ri]].pt)>0) {
				// convex corner: cut if off
				op[_op_idx][0] = rc[ri - 1];
				op[_op_idx][1] = rc[ri];
				op[_op_idx][2] = v;
				_op_idx++;
				ri--;
			} else {	// non-convex add v to the chain
				ri++;
				rc[ri] = v;
				vpos = _mchain[vpos].next;
				v = _mchain[vpos].vnum;
			}
		} else { // reflex-chain empty: add v to the
				 // reflex chain and advance it
			rc[++ri] = v;
			vpos = _mchain[vpos].next;
			v = _mchain[vpos].vnum;
		}
	} // end-while

	// reached the bottom vertex. Add in the triangle formed
	op[_op_idx][0] = rc[ri - 1];
	op[_op_idx][1] = rc[ri];
	op[_op_idx][2] = v;
	_op_idx++;
	ri--;

	return 0;
} // TriangulatablePolygon::triangulate_single_polygon()

